"""Python SDK for Finergy MIA POS eComm API"""

import logging
import time

from .finergy_mia_pos_common import FinergyMiaPosCommon, FinergyClientApiException

logger = logging.getLogger(__name__)


class FinergyMiaPosAuthClient:
    """
    Handles authentication with the MIA POS eComm API.
    Provides methods to generate, refresh, and retrieve access tokens.

    https://github.com/finergy-tech/mia-pay-ecomm-integration/blob/main/docs/mia-ecomm-api_v0.0.1.html
    """

    AUTH_TOKEN = '/ecomm/api/v1/token'
    AUTH_TOKEN_REFRESH = '/ecomm/api/v1/token/refresh'

    _base_url: str = None
    _merchant_id: str = None
    _secret_key: str = None

    _access_token: str = None
    _refresh_token: str = None
    _access_expire_time: str = None

    def __init__(self, base_url: str, merchant_id: str, secret_key: str):
        self._base_url = base_url.rstrip('/')
        self._merchant_id = merchant_id
        self._secret_key = secret_key

    #region Get access token
    def get_access_token(self):
        """
        Retrieves the current access token.

        If the current access token is valid, it will return the cached token.
        Otherwise, it will attempt to refresh the token or generate a new one.

        Returns:
            str: The valid access token.

        Raises:
            FinergyClientApiException: If the token cannot be generated or refreshed.
        """

        if self._access_token and not self._is_token_expired():
            return self._access_token

        if self._refresh_token:
            try:
                return self._refresh_access_token()
            except Exception:
                logger.exception('MIA POS refresh token failed')

        return self._generate_new_tokens()

    async def get_access_token_async(self):
        """
        Retrieves the current access token.

        If the current access token is valid, it will return the cached token.
        Otherwise, it will attempt to refresh the token or generate a new one.

        Returns:
            str: The valid access token.

        Raises:
            FinergyClientApiException: If the token cannot be generated or refreshed.
        """

        if self._access_token and not self._is_token_expired():
            return self._access_token

        if self._refresh_token:
            try:
                return await self._refresh_access_token_async()
            except Exception:
                logger.exception('MIA POS refresh token failed')

        return await self._generate_new_tokens_async()
    #endregion

    #region Generate new auth tokens API
    def _generate_new_tokens(self):
        """
        Generates a new access token using the merchant credentials.
        Sends a request to the MIA POS API to obtain a new access and refresh token pair.

        Returns:
            str: The newly generated access token.

        Raises:
            FinergyClientApiException: If the API request fails or no access token is returned.
        """

        url = self._base_url + self.AUTH_TOKEN
        generate_tokens_data = self._build_generate_tokens_data()

        response = FinergyMiaPosCommon.send_request(method='POST', url=url, data=generate_tokens_data)
        self._parse_response_token(response)

        if not self._access_token:
            raise FinergyClientApiException(f'Failed to retrieve access token by merchantId {self._merchant_id}. accessToken is missing from the response')

        return self._access_token

    async def _generate_new_tokens_async(self):
        """
        Generates a new access token using the merchant credentials.
        Sends a request to the MIA POS API to obtain a new access and refresh token pair.

        Returns:
            str: The newly generated access token.

        Raises:
            FinergyClientApiException: If the API request fails or no access token is returned.
        """

        url = self._base_url + self.AUTH_TOKEN
        generate_tokens_data = self._build_generate_tokens_data()

        response = await FinergyMiaPosCommon.send_request_async(method='POST', url=url, data=generate_tokens_data)
        self._parse_response_token(response)

        if not self._access_token:
            raise FinergyClientApiException(f'Failed to retrieve access token by merchantId {self._merchant_id}. accessToken is missing from the response')

        return self._access_token

    def _build_generate_tokens_data(self):
        tokens_data = {
            'merchantId': self._merchant_id,
            'secretKey': self._secret_key,
        }

        return tokens_data
    #endregion

    #region Refresh auth tokens API
    def _refresh_access_token(self):
        """
        Refreshes the current access token using the refresh token.
        Sends a request to the MIA POS API to refresh the access token.

        Returns:
            str: The refreshed access token.

        Raises:
            FinergyClientApiException: If the API request fails or no access token is returned.
        """

        url = self._base_url + self.AUTH_TOKEN_REFRESH
        refresh_tokens_data = self._build_refresh_tokens_data()

        response = FinergyMiaPosCommon.send_request(method='POST', url=url, data=refresh_tokens_data)
        self._parse_response_token(response)

        if not self._access_token:
            raise FinergyClientApiException(f'Failed to refresh access token by merchantId {self._merchant_id}. accessToken is missing from the response')

        return self._access_token

    async def _refresh_access_token_async(self):
        """
        Refreshes the current access token using the refresh token.
        Sends a request to the MIA POS API to refresh the access token.

        Returns:
            str: The refreshed access token.

        Raises:
            FinergyClientApiException: If the API request fails or no access token is returned.
        """

        url = self._base_url + self.AUTH_TOKEN_REFRESH
        refresh_tokens_data = self._build_refresh_tokens_data()

        response = await FinergyMiaPosCommon.send_request_async(method='POST', url=url, data=refresh_tokens_data)
        self._parse_response_token(response)

        if not self._access_token:
            raise FinergyClientApiException(f'Failed to refresh access token by merchantId {self._merchant_id}. accessToken is missing from the response')

        return self._access_token

    def _build_refresh_tokens_data(self):
        tokens_data = {
            'refreshToken': self._refresh_token
        }

        return tokens_data
    #endregion

    def _is_token_expired(self):
        """
        Checks whether the current access token has expired.

        Returns:
            bool: True if the token is expired; otherwise, False.
        """

        return not self._access_expire_time or time.time() >= self._access_expire_time

    def _parse_response_token(self, response: dict):
        """
        Parses the token response from the MIA POS API.
        Extracts the access token, refresh token, and token expiration time from the response.

        Args:
            response (dict): The decoded API response containing token details.
        """

        self._access_token = response.get('accessToken')
        self._refresh_token = response.get('refreshToken')
        expires_in: int = response.get('accessTokenExpiresIn', 0)
        self._access_expire_time = time.time() + expires_in - 10



"""Python SDK for Finergy MIA POS eComm API"""

import logging
import httpx

logger = logging.getLogger(__name__)


class FinergyMiaPosCommon:
    DEFAULT_TIMEOUT = 30

    @classmethod
    def send_request(cls, method: str, url: str, data: dict = None, params: dict = None, token: str = None):
        """
        Sends an HTTP request to the MIA POS API.

        Args:
            method (str): HTTP method (e.g., 'POST', 'GET').
            url (str): The API endpoint URL.
            data (dict): The request payload (for POST requests).
            params (dict): Request URL params.
            token: Access token for authorization (optional).

        Returns:
            dict: The decoded JSON response from the API.

        Raises:
            FinergyClientApiException: If a network error, HTTP error, or JSON decoding failure occurs.
        """

        try:
            auth = BearerAuth(token) if token else None

            logger.debug(f'{cls.__qualname__} Request: %s %s', method, url, extra={'method': method, 'url': url, 'data': data, 'params': params, 'token': token})
            with httpx.Client() as client:
                response = client.request(method=method, url=url, params=params, json=data, auth=auth, timeout=cls.DEFAULT_TIMEOUT)
                return cls._process_response(response=response)

        except Exception as ex:
            logger.exception(cls.__qualname__)
            raise FinergyClientApiException(f'MIA POS client url {url}, method {method} error: {ex}') from ex

    @classmethod
    async def send_request_async(cls, method: str, url: str, data: dict = None, params: dict = None, token: str = None):
        """
        Sends an async HTTP request to the MIA POS API.

        Args:
            method (str): HTTP method (e.g., 'POST', 'GET').
            url (str): The API endpoint URL.
            data (dict): The request payload (for POST requests).
            params (dict): Request URL params.
            token: Access token for authorization (optional).

        Returns:
            dict: The decoded JSON response from the API.

        Raises:
            FinergyClientApiException: If a network error, HTTP error, or JSON decoding failure occurs.
        """

        try:
            auth = BearerAuth(token) if token else None

            logger.debug(f'{cls.__qualname__} Request: %s %s', method, url, extra={'method': method, 'url': url, 'data': data, 'params': params, 'token': token})
            async with httpx.AsyncClient() as client:
                response = await client.request(method=method, url=url, params=params, json=data, auth=auth, timeout=cls.DEFAULT_TIMEOUT)
                return cls._process_response(response=response)

        except Exception as ex:
            logger.exception(cls.__qualname__)
            raise FinergyClientApiException(f'MIA POS client url {url}, method {method} error: {ex}') from ex

    @classmethod
    def _process_response(cls, response: httpx.Response):
        if response.is_error:
            logger.error(f'{cls.__qualname__} Error: %d %s', response.status_code, response.text, extra={'method': response.request.method, 'url': response.request.url, 'params': response.request.url.params, 'response_text': response.text, 'status_code': response.status_code})
            #response.raise_for_status()
            raise FinergyClientApiException(f'MIA POS client url {response.request.url}, method {response.request.method} HTTP Error: {response.status_code}, Response: {response.text}')

        response_json: dict = response.json()
        logger.debug(f'{cls.__qualname__} Response: %d %s %s', response.status_code, response.request.method, response.request.url, extra={'method': response.request.method, 'url': response.request.url, 'params': response.request.url.params, 'response_json': response_json, 'status_code': response.status_code})
        return response_json

#region Auth
class BearerAuth(httpx.Auth):
    """Attaches HTTP Bearer Token Authentication to the given Request object."""
    # https://www.python-httpx.org/advanced/authentication/#custom-authentication-schemes

    token: str = None

    def __init__(self, token: str):
        self.token = token

    def auth_flow(self, request: httpx.Request):
        request.headers['Authorization'] = f'Bearer {self.token}'
        yield request
#endregion

#region Exceptions
class FinergyMiaPosSdkException(Exception):
    """Base exception class for MIA POS SDK."""
    pass

class FinergyClientApiException(FinergyMiaPosSdkException):
    """Represents an exception thrown when an API request fails."""

    _http_status_code: int = None
    _error_code: str = None
    _error_message: str = None

    def __init__(self, message: str, http_status_code: int = None, error_code: str = None, error_message: str = None):
        """
        FinergyClientApiException constructor.

        Args:
            message (str): General error message.
            http_status_code (int): HTTP status code.
            error_code (str): Error code returned by the API.
            error_message (str): Error message returned by the API.
        """
        super().__init__(message)
        self._http_status_code = http_status_code
        self._error_code = error_code
        self._error_message = error_message

    def get_http_status_code(self):
        """
        Get the HTTP status code.

        Returns:
            int: HTTP status code or None
        """

        return self._http_status_code

    def get_error_code(self):
        """
        Get the error code returned by the API.

        Returns:
            str: Error code or None
        """

        return self._error_code

    def get_error_message(self):
        """
        Get the error message returned by the API.

        Returns:
            str: Error message or None
        """

        return self._error_message

class FinergyValidationException(FinergyMiaPosSdkException):
    """Represents an exception thrown when validation of input data fails."""

    _invalid_fields: list = None

    def __init__(self, message: str, invalid_fields: list = None):
        """
        ValidationException constructor.

        Args:
            message (str): Error message
            invalid_fields (list): List of invalid fields
        """

        super().__init__(message)
        self._invalid_fields = invalid_fields or []

    def get_invalid_fields(self):
        """
        Get the invalid fields.

        Returns:
            list: List of invalid field names
        """

        return self._invalid_fields
#endregion



"""Python SDK for Finergy MIA POS eComm API"""

from .finergy_mia_pos_common import FinergyMiaPosCommon, FinergyClientApiException


class FinergyMiaPosApiClient:
    """
    Handles API requests to the MIA POS eComm API.
    Provides methods for creating payments, checking payment status,
    and retrieving the public key.

    https://github.com/finergy-tech/mia-pay-ecomm-integration/blob/main/docs/mia-ecomm-api_v0.0.1.html
    """

    API_PAY = '/ecomm/api/v1/pay'
    API_PAYMENT = '/ecomm/api/v1/payment'
    API_PUBLIC_KEY = '/ecomm/api/v1/public-key'

    _base_url: str = None

    def __init__(self, base_url: str):
        self._base_url = base_url.rstrip('/')

    #region Create payment
    def create_payment(self, token: str, payment_data: dict):
        """
        Creates a new payment.
        Sends a POST request to the MIA POS API to create a payment.

        Args:
            token (str): Access token for authorization.
            payment_data (dict): A dict containing payment details by miaEcomm protocol

        Returns:
            dict: Response from the API containing payment details.

        Raises:
            FinergyClientApiException: If the API request fails or returns an error.
        """

        url = self._base_url + self.API_PAY
        return FinergyMiaPosCommon.send_request(method='POST', url=url, data=payment_data, token=token)

    async def create_payment_async(self, token: str, payment_data: dict):
        """
        Creates a new payment.
        Sends a POST request to the MIA POS API to create a payment.

        Args:
            token (str): Access token for authorization.
            payment_data (dict): A dict containing payment details by miaEcomm protocol

        Returns:
            dict: Response from the API containing payment details.

        Raises:
            FinergyClientApiException: If the API request fails or returns an error.
        """

        url = self._base_url + self.API_PAY
        return await FinergyMiaPosCommon.send_request_async(method='POST', url=url, data=payment_data, token=token)
    #endregion

    #region Payment status
    def get_payment_status(self, token: str, payment_id: str):
        """
        Retrieves the status of a payment.
        Sends a GET request to the MIA POS API to retrieve the payment status by its ID.

        Args:
            token (str): Access token for authorization.
            payment_id (str): Unique identifier of the payment.

        Returns:
            dict: Response from the API containing the payment status.

        Raises:
            FinergyClientApiException: If the API request fails or returns an error.
        """

        url = self._base_url + f'{self.API_PAYMENT}/{payment_id}'
        return FinergyMiaPosCommon.send_request(method='GET', url=url, token=token)

    async def get_payment_status_async(self, token: str, payment_id: str):
        """
        Retrieves the status of a payment.
        Sends a GET request to the MIA POS API to retrieve the payment status by its ID.

        Args:
            token (str): Access token for authorization.
            payment_id (str): Unique identifier of the payment.

        Returns:
            dict: Response from the API containing the payment status.

        Raises:
            FinergyClientApiException: If the API request fails or returns an error.
        """

        url = self._base_url + f'{self.API_PAYMENT}/{payment_id}'
        return await FinergyMiaPosCommon.send_request_async(method='GET', url=url, token=token)
    #endregion

    #region Public key
    def get_public_key(self, token: str):
        """
        Retrieves the public key from the MIA POS API.
        Sends a GET request to retrieve the public key for signature verification.

        Args:
            token (str): Access token for authorization.

        Returns:
            str: The public key returned by the API.

        Raises:
            FinergyClientApiException: If the public key is not found or the API request fails.
        """

        url = self._base_url + self.API_PUBLIC_KEY
        response = FinergyMiaPosCommon.send_request(method='GET', url=url, token=token)

        public_key: str = response.get('publicKey') if response else None

        if not public_key:
            raise FinergyClientApiException('Public key not found in the response')

        return public_key

    async def get_public_key_async(self, token: str):
        """
        Retrieves the public key from the MIA POS API.
        Sends a GET request to retrieve the public key for signature verification.

        Args:
            token (str): Access token for authorization.

        Returns:
            str: The public key returned by the API.

        Raises:
            FinergyClientApiException: If the public key is not found or the API request fails.
        """

        url = self._base_url + self.API_PUBLIC_KEY
        response = await FinergyMiaPosCommon.send_request_async(method='GET', url=url, token=token)

        public_key: str = response.get('publicKey') if response else None

        if not public_key:
            raise FinergyClientApiException('Public key not found in the response')

        return public_key
    #endregion



"""Python SDK for Finergy MIA POS eComm API"""

import logging

from .finergy_mia_pos_sdk import FinergyMiaPosSdk
from .finergy_mia_pos_auth_client import FinergyMiaPosAuthClient
from .finergy_mia_pos_api_client import FinergyMiaPosApiClient
from .finergy_mia_pos_common import FinergyValidationException, FinergyClientApiException

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())



"""Python SDK for Finergy MIA POS eComm API"""

import logging
import base64

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.exceptions import InvalidSignature

from .finergy_mia_pos_auth_client import FinergyMiaPosAuthClient
from .finergy_mia_pos_api_client import FinergyMiaPosApiClient
from .finergy_mia_pos_common import FinergyValidationException

logger = logging.getLogger(__name__)


# Based on Finergy MIA POS PHP SDK https://github.com/finergy-tech/mia-pay-ecomm-php-sdk (https://packagist.org/packages/finergy/mia-pos-sdk)

class FinergyMiaPosSdk:
    """
    * Integration with MIA POS eComm https://github.com/finergy-tech/mia-pay-ecomm-integration
    * MIA POS eComm API https://github.com/finergy-tech/mia-pay-ecomm-integration/blob/main/docs/mia-ecomm-api_v0.0.1.html
    """

    # https://github.com/finergy-tech/mia-pay-ecomm-integration/blob/main/docs/en/protocol-overview.md#1-initial-configuration
    TEST_BASE_URL = 'https://ecomm-test.miapos.md'

    _api_client = None
    _auth_client = None

    def __init__(self, base_url: str, merchant_id: str, secret_key: str):
        """
        Initializes the SDK with the base URL, Merchant ID, and Secret Key.

        Args:
            base_url (str): Base URL for the MIA POS API.
            merchant_id (str): Merchant ID provided by MIA POS.
            secret_key (str): Secret Key for authentication.

        Raises:
            FinergyValidationException: If any of the required parameters are missing.
        """

        if not base_url:
            raise FinergyValidationException('Base URL is required.')

        if not merchant_id:
            raise FinergyValidationException('Merchant ID is required.')

        if not secret_key:
            raise FinergyValidationException('Secret Key is required.')

        self._auth_client = FinergyMiaPosAuthClient(base_url=base_url, merchant_id=merchant_id, secret_key=secret_key)
        self._api_client = FinergyMiaPosApiClient(base_url=base_url)

    #region Create payment
    def create_payment(self, payment_data: dict):
        """
        Creates a new payment.

        Args:
            payment_data (dict): A dict containing payment details by miaEcomm protocol.
                Required keys: 'terminalId', 'orderId', 'amount', 'currency', 'payDescription'.

        Returns:
            dict: Response from the API.

        Raises:
            FinergyValidationException: If required parameters are missing.
            FinergyClientApiException: If there is an API error during the request.
        """

        required_fields = ['terminalId', 'orderId', 'amount', 'currency', 'payDescription']

        self._validate_parameters(data=payment_data, required_fields=required_fields)
        token = self._get_access_token()
        return self._api_client.create_payment(token=token, payment_data=payment_data)

    async def create_payment_async(self, payment_data: dict):
        """
        Creates a new payment.

        Args:
            payment_data (dict): A dict containing payment details by miaEcomm protocol.
                Required keys: 'terminalId', 'orderId', 'amount', 'currency', 'payDescription'.

        Returns:
            dict: Response from the API.

        Raises:
            FinergyValidationException: If required parameters are missing.
            FinergyClientApiException: If there is an API error during the request.
        """

        required_fields = ['terminalId', 'orderId', 'amount', 'currency', 'payDescription']

        self._validate_parameters(data=payment_data, required_fields=required_fields)
        token = await self._get_access_token_async()
        return await self._api_client.create_payment_async(token=token, payment_data=payment_data)
    #endregion

    #region Get payment status
    def get_payment_status(self, payment_id: str):
        """
        Retrieves the status of a payment by its ID.

        Args:
            payment_id (str): The unique payment ID.

        Returns:
            dict: Response from the API.

        Raises:
            FinergyValidationException: If the payment ID is empty.
            FinergyClientApiException: If there is an API error during the request.
        """

        if not payment_id:
            raise FinergyValidationException('Payment ID is required.')

        token = self._get_access_token()
        return self._api_client.get_payment_status(token=token, payment_id=payment_id)

    async def get_payment_status_async(self, payment_id: str):
        """
        Retrieves the status of a payment by its ID.

        Args:
            payment_id (str): The unique payment ID.

        Returns:
            dict: Response from the API.

        Raises:
            FinergyValidationException: If the payment ID is empty.
            FinergyClientApiException: If there is an API error during the request.
        """

        if not payment_id:
            raise FinergyValidationException('Payment ID is required.')

        token = await self._get_access_token_async()
        return await self._api_client.get_payment_status_async(token=token, payment_id=payment_id)
    #endregion

    #region Callback signature validation
    def validate_callback_signature(self, callback_data: dict):
        """
        Validates the callback data signature.

        Args:
            callback_data (dict): Callback payload data.

        Returns:
            bool: Callback data payload signature validation result.

        Raises:
            FinergyValidationException: If required parameters are missing or fails to verify callback signature.
            FinergyClientApiException: If there is an API error during the request.
        """

        # https://github.com/finergy-tech/mia-pay-ecomm-integration/blob/main/docs/en/signature-verification.md
        # https://github.com/finergy-tech/mia-pay-ecomm-php-sdk/blob/main/examples/process_callback.php
        callback_signature: str = callback_data.get('signature')
        callback_result: dict[str, any] = callback_data.get('result')

        if not callback_signature or not callback_result:
            raise FinergyValidationException('Missing result or signature in callback data.')

        sign_string = self.form_sign_string_by_result(result_data=callback_result)
        validation_result = self.verify_signature(result_str=sign_string, signature=callback_signature)

        return validation_result

    async def validate_callback_signature_async(self, callback_data: dict):
        """
        Validates the callback data signature.

        Args:
            callback_data (dict): Callback payload data.

        Returns:
            bool: Callback data payload signature validation result.

        Raises:
            FinergyValidationException: If required parameters are missing or fails to verify callback signature.
            FinergyClientApiException: If there is an API error during the request.
        """

        callback_signature: str = callback_data.get('signature')
        callback_result: dict[str, any] = callback_data.get('result')

        if not callback_signature or not callback_result:
            raise FinergyValidationException('Missing result or signature in callback data.')

        sign_string = self.form_sign_string_by_result(result_data=callback_result)
        validation_result = await self.verify_signature_async(result_str=sign_string, signature=callback_signature)

        return validation_result

    def verify_signature(self, result_str: str, signature: str):
        """
        Verifies the signature of a payment result.

        Args:
            result_str (str): The result string to verify.
            signature (str): The provided signature.

        Returns:
            bool: True if the signature is valid; otherwise, False.

        Raises:
            FinergyValidationException: If required parameters are missing.
            FinergyClientApiException: If there is an API error during the request.
        """

        if not result_str:
            raise FinergyValidationException('Result string is required.')

        if not signature:
            raise FinergyValidationException('Signature is required.')

        token = self._get_access_token()
        public_key_str = self._api_client.get_public_key(token=token)

        if not public_key_str:
            raise FinergyValidationException('Public key is missing in the response.')

        try:
            return self._verify_result_signature(
                result_str=result_str,
                signature=signature,
                public_key_str=public_key_str)

        except Exception as ex:
            logger.exception(self.__class__.__qualname__)
            raise FinergyValidationException(f'Failed to verify signature: {ex}') from ex

    async def verify_signature_async(self, result_str: str, signature: str):
        """
        Verifies the signature of a payment result.

        Args:
            result_str (str): The result string to verify.
            signature (str): The provided signature.

        Returns:
            bool: True if the signature is valid; otherwise, False.

        Raises:
            FinergyValidationException: If required parameters are missing.
            FinergyClientApiException: If there is an API error during the request.
        """

        if not result_str:
            raise FinergyValidationException('Result string is required.')

        if not signature:
            raise FinergyValidationException('Signature is required.')

        token = await self._get_access_token_async()
        public_key_str = await self._api_client.get_public_key_async(token=token)

        if not public_key_str:
            raise FinergyValidationException('Public key is missing in the response.')

        try:
            return self._verify_result_signature(
                result_str=result_str,
                signature=signature,
                public_key_str=public_key_str)

        except Exception as ex:
            logger.exception(self.__class__.__qualname__)
            raise FinergyValidationException(f'Failed to verify signature: {ex}') from ex

    @staticmethod
    def form_sign_string_by_result(result_data: dict[str, any]):
        """
        Forms a signature string based on result data.

        Args:
            result_data (dict): A set of data received when receiving the payment status or when receiving the payment result on callbackUrl.

        Returns:
            str: Generated string for the data set, for signature verification.

        Raises:
            FinergyValidationException: If the result data is invalid.
        """

        if not result_data:
            raise FinergyValidationException('Result data must be a non-empty dict.')

        sorted_items = sorted(result_data.items())

        formatted_values = []
        for key, value in sorted_items:
            if key == 'amount':
                formatted_values.append(f'{float(value):.2f}')
            else:
                formatted_values.append(str(value))

        return ';'.join(formatted_values)

    @classmethod
    def _verify_result_signature(cls, result_str: str, signature: str, public_key_str: str):
        public_key_pem = (
            '-----BEGIN PUBLIC KEY-----\n' +
            '\n'.join(public_key_str[i:i+64] for i in range(0, len(public_key_str), 64)) +
            '\n-----END PUBLIC KEY-----'
        )

        public_key = serialization.load_pem_public_key(public_key_pem.encode())
        decoded_signature = base64.b64decode(signature)

        try:
            public_key.verify(
                decoded_signature,
                result_str.encode(),
                padding.PKCS1v15(),
                hashes.SHA256())

            return True
        except InvalidSignature:
            return False
    #endregion

    #region Validation
    def _validate_parameters(self, data: dict, required_fields: list[str]):
        """
        Validate the required parameters.

        Args:
            data (dict): Input data to validate.
            required_fields (list): List of required fields.

        Raises:
            ValidationException: If a required field is missing.
        """

        missing_fields = []

        for field in required_fields:
            if field not in data or not data[field]:
                missing_fields.append(field)

        if missing_fields:
            raise FinergyValidationException(
                f'Missing required fields: {", ".join(missing_fields)}',
                invalid_fields=missing_fields)
    #endregion

    #region Access token
    def _get_access_token(self):
        return self._auth_client.get_access_token()

    async def _get_access_token_async(self):
        return await self._auth_client.get_access_token_async()
    #endregion
